from inspect import Parameter, isawaitable, signature
from itertools import accumulate
from typing import AsyncIterable, Awaitable, Callable, Iterable, TypeVar, cast

T = TypeVar("T")


def appender(to_append: list[T]) -> Callable[[T], T]:
    def append_processer(func: T) -> T:
        to_append.append(func)

        return func

    return append_processer


def is_positional_parameter(p: Parameter):
    return p.kind is Parameter.POSITIONAL_OR_KEYWORD or p.kind is Parameter.KEYWORD_ONLY


def count_position_parameters(func):
    return sum(map(is_positional_parameter, signature(func).parameters.values()))


async def resolve(maybe_awaitable: T | Awaitable[T], /) -> T:
    while isawaitable(maybe_awaitable):
        maybe_awaitable = await maybe_awaitable

    return maybe_awaitable  # type: ignore


async def async_accumulate(async_iterable: AsyncIterable[str]):
    result = ""
    async for delta in async_iterable:
        result += delta
        yield result


def accumulate_any(any_iterable: Iterable[str] | AsyncIterable[str]):
    if "__aiter__" in dir(any_iterable):
        return async_accumulate(any_iterable)  # type: ignore

    async def _():
        for i in accumulate(cast(Iterable[str], any_iterable)):
            yield i

    return _()



from functools import cache, cached_property, wraps
from inspect import currentframe, isclass
from pathlib import Path
from re import compile
from typing import Any, Callable, ParamSpec, TypeVar

split_template_tokens = compile(
    r"((?:\s{%-|{%).*?(?:%}|-%}\s))|((?:\s{{-|{{)[\s\S]*?(?:}}|-}}\s))|((?:\s{#-|{#)[\s\S]*?(?:#}|-#}\s))"
).split


var_name_checker = compile(r"[_a-zA-Z]\w*$")

is_message_start = compile(r"<\|\s?(user|system|assistant)\s?(\w{1,64})?\s?\|>")


def is_not_valid(name: str):
    return not var_name_checker.match(name)


def ensure_valid(name):
    if is_not_valid(name):
        raise NameError(name)


class AutoNaming:
    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        obj._bind_frame()
        return obj

    def _bind_frame(self):
        self._frame = currentframe()

    @cached_property
    def _name(self):
        f = self._frame
        if f and f.f_back and (frame := f.f_back.f_back):
            for name, var in frame.f_locals.items():
                if var is self:
                    return name

    @property
    def class_name(self):
        return self.__class__.__name__

    fallback_name = class_name

    @property
    def name(self):
        return self._name or self.fallback_name

    @name.setter
    def name(self, name):
        self._name = name
        self._frame = None

    @name.deleter
    def name(self):
        del self._name

    def __repr__(self):
        if self._name:
            return f"<{self.class_name} {self.name}>"
        else:
            return f"<{self.class_name}>"

    def __str__(self):
        return f"<{self.name}>"


P = ParamSpec("P")
T = TypeVar("T")


def cache_once(func: Callable[P, T]) -> Callable[P, T]:
    result = None

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        nonlocal result
        if result is None:
            result = func(*args, **kwargs)
        return result

    return wrapper


@cache_once
def get_builtins() -> dict[str, Any]:
    return __builtins__ if isinstance(__builtins__, dict) else __builtins__.__dict__


@cache
def version(package: str):
    from importlib.metadata import PackageNotFoundError, version

    try:
        return version(package)
    except PackageNotFoundError:
        return None


@cache_once
def get_user_agent(self, *additional_packages: tuple[str, str]):
    from sys import version as py_version

    return " ".join(
        (
            f"Promplate/{version('promplate')} ({self.__name__ if isclass(self) else self.__class__.__name__})",
            *(f"{name}/{v}" for name, v in additional_packages),
            f"HTTPX/{version('httpx') or '-'}",
            f"Python/{py_version.split()[0]}",
        )
    )


@cache_once
def _is_http2_available():
    try:
        import h2  # type: ignore

        return True
    except ImportError:
        return False


@cache_once
def _get_client():
    from httpx import Client

    return Client(follow_redirects=True, http2=_is_http2_available())


@cache_once
def _get_aclient():
    from httpx import AsyncClient

    return AsyncClient(follow_redirects=True, http2=_is_http2_available())


def add_linecache(filename: str, source_getter: Callable[[], str]):
    import linecache

    linecache.cache[filename] = (source_getter,)


def save_tempfile(filename: str, source: str, auto_deletion: bool):
    from tempfile import gettempdir, mkdtemp

    for i in Path(gettempdir()).glob(f"promplate-*/{filename}"):
        i.unlink()

    file = Path(mkdtemp(prefix="promplate-")) / filename
    file.write_text(source)

    if auto_deletion:
        import atexit

        @atexit.register
        def _():
            file.unlink(missing_ok=True)
            file.parent.rmdir()

    return file



from copy import copy
from functools import cached_property
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Callable, ParamSpec, TypeVar

from openai import AsyncClient, Client  # type: ignore

from ...prompt.chat import Message, ensure
from ...prompt.utils import _get_aclient, _get_client, get_user_agent
from ..base import *

P = ParamSpec("P")
T = TypeVar("T")


class Config(Configurable):
    def __init__(self, **config):
        super().__init__(**config)
        self._run_config = {}

    def bind(self, **run_config):
        obj = copy(self)
        obj._run_config = self._run_config | run_config
        return obj

    @cached_property
    def _user_agent(self):
        from openai.version import VERSION

        return get_user_agent(self, ("OpenAI", VERSION))

    @property
    def _config(self):  # type: ignore
        ua_header = {"User-Agent": self._user_agent}
        config = dict(super()._config)
        config["default_headers"] = config.get("default_headers", {}) | ua_header
        return MappingProxyType(config)

    @cached_property
    def _client(self):
        if "http_client" in self._config:
            return Client(**self._config)
        else:
            return Client(**self._config, http_client=_get_client())

    @cached_property
    def _aclient(self):
        if "http_client" in self._config:
            return AsyncClient(**self._config)
        else:
            return AsyncClient(**self._config, http_client=_get_aclient())


if TYPE_CHECKING:

    def same_params_as(_: Callable[P, Any]) -> Callable[[Callable[..., None]], Callable[P, None]]: ...

    class ClientConfig(Config):
        @same_params_as(Client)
        def __init__(self, **config): ...

    class AsyncClientConfig(Config):
        @same_params_as(AsyncClient)
        def __init__(self, **config): ...

else:
    ClientConfig = AsyncClientConfig = Config


class TextComplete(ClientConfig):
    def __call__(self, text: str, /, **config):
        config = self._run_config | config | {"stream": False, "prompt": text}
        result = self._client.completions.create(**config)
        return result.choices[0].text


class AsyncTextComplete(AsyncClientConfig):
    async def __call__(self, text: str, /, **config):
        config = self._run_config | config | {"stream": False, "prompt": text}
        result = await self._aclient.completions.create(**config)
        return result.choices[0].text


class TextGenerate(ClientConfig):
    def __call__(self, text: str, /, **config):
        config = self._run_config | config | {"stream": True, "prompt": text}
        stream = self._client.completions.create(**config)
        for event in stream:
            try:
                yield event.choices[0].text
            except AttributeError:
                pass


class AsyncTextGenerate(AsyncClientConfig):
    async def __call__(self, text: str, /, **config):
        config = self._run_config | config | {"stream": True, "prompt": text}
        stream = await self._aclient.completions.create(**config)
        async for event in stream:
            try:
                yield event.choices[0].text
            except AttributeError:
                pass


class ChatComplete(ClientConfig):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._run_config | config | {"stream": False, "messages": messages}
        result = self._client.chat.completions.create(**config)
        return result.choices[0].message.content


class AsyncChatComplete(AsyncClientConfig):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._run_config | config | {"stream": False, "messages": messages}
        result = await self._aclient.chat.completions.create(**config)
        return result.choices[0].message.content


class ChatGenerate(ClientConfig):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._run_config | config | {"stream": True, "messages": messages}
        stream = self._client.chat.completions.create(**config)
        for event in stream:
            try:
                yield event.choices[0].delta.content or ""
            except AttributeError:
                pass


class AsyncChatGenerate(AsyncClientConfig):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._run_config | config | {"stream": True, "messages": messages}
        stream = await self._aclient.chat.completions.create(**config)
        async for event in stream:
            try:
                yield event.choices[0].delta.content or ""
            except AttributeError:
                pass


class SyncTextOpenAI(ClientConfig, LLM):
    complete = TextComplete.__call__  # type: ignore
    generate = TextGenerate.__call__  # type: ignore


class AsyncTextOpenAI(AsyncClientConfig, LLM):
    complete = AsyncTextComplete.__call__  # type: ignore
    generate = AsyncTextGenerate.__call__  # type: ignore


class SyncChatOpenAI(ClientConfig, LLM):
    complete = ChatComplete.__call__  # type: ignore
    generate = ChatGenerate.__call__  # type: ignore


class AsyncChatOpenAI(AsyncClientConfig, LLM):
    complete = AsyncChatComplete.__call__  # type: ignore
    generate = AsyncChatGenerate.__call__  # type: ignore


__all__ = (
    "TextComplete",
    "AsyncTextComplete",
    "TextGenerate",
    "AsyncTextGenerate",
    "ChatComplete",
    "AsyncChatComplete",
    "ChatGenerate",
    "AsyncChatGenerate",
    "SyncTextOpenAI",
    "AsyncTextOpenAI",
    "SyncChatOpenAI",
    "AsyncChatOpenAI",
)
